// Ardrone.cpp
//
// Remote Control module for AR.Drone
/// \author  Mike McCauley (mikem@airspayce.com)
///
// Copyright (C) 2011 Mike McCauley
// $Id: Ardrone.cpp,v 1.1 2011/03/16 23:46:12 mikem Exp mikem $

#if ARDUINO >= 100
#include <Arduino.h>
#else
#include "WProgram.h"
#endif
#include "Ardrone.h"
#include "g2100.h"
#include <stdio.h>

// Command buffer for accumulating commands
char cmd[ARDRONE_MAX_AT_COMMAND_LENGTH];

// There is precisely one Ardrone instance, which we remember here
static Ardrone* ardrone;

// WiShield Wireless configuration parameters
// AdHoc connection for direct connection to AR.Drone
// The ssid must be defined in the parent sketch
unsigned char local_ip[] = {192,168,1,2};              // My address
unsigned char gateway_ip[] = {169,254,1,100};
unsigned char subnet_mask[] = {255,255,255,0};
unsigned char security_type = ZG_SECURITY_TYPE_NONE;	// _NONE, _WEP, _WPA, _WPA2
unsigned char wireless_mode = WIRELESS_MODE_ADHOC;

// WPA/WPA2 passphrase (if wireless_mode ZG_SECURITY_TYPE_WPA or ZG_SECURITY_TYPE_WPA2)
const prog_char security_passphrase[] PROGMEM = {"notrequired"};	// max 64 characters

// WEP 128-bit keys if wireless_mode ZG_SECURITY_TYPE_WEP
// sample HEX keys
prog_uchar wep_keys[] PROGMEM = 
{ 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d,	// Key 0
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	// Key 1
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	// Key 2
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00	// Key 3
};

unsigned char ssid_len;
unsigned char security_passphrase_len;
// End WiShield Wireless configuration parameters 

/////////////////////////////////////////////////////////////////////
// Glue between the WiShield C side and the Ardrone C++ side
// Most of this is equivalent to the contents of udpapp.c example in WiShield
// Caution: need to edit WiShield headers in order to enable UDP apps.
extern "C"
{
#include "uip.h"
#include "udpapp.h"
#include "config.h"

    static struct udpapp_state s;

    // REVISIT: should be configurable from outside here
    void udpapp_init(void)
    {
	uip_ipaddr_t addr;
	struct uip_udp_conn *udp;
	struct uip_conn *tcp;

	uip_ipaddr(&addr, 192,168,1,1); // Dest IP Address
	udp = uip_udp_new(&addr, HTONS(ARDRONE_AT_PORT)); // UDP
	tcp = uip_connect(&addr, HTONS(ARDRONE_CONFIGDATA_PORT)); // TCP
	PT_INIT(&s.pt);
    }

    static PT_THREAD(handle_udp_connection(void))
    {
	PT_BEGIN(&s.pt);
	ardrone->periodicTask();
	PT_END(&s.pt);
    }

    void debug(char* m)
    {
	Serial.println(m);
    }

    // This is called by the WiShield UIP code for events on TCP connections
    // The strange name is an artifact due to using the UDP application
    void dummy_app_appcall(void)
    {
	if (uip_connected())
	    ardrone->connected();
    }

    // This is called by the WiShield UIP code for events on UDP connections
    void udpapp_appcall(void)
    {
	if (uip_poll())
	    handle_udp_connection();
    }
}

/////////////////////////////////////////////////////////////////////
Ardrone::Ardrone()
{
    // record this as the singleton
    ardrone = this;
    _state = ARDRONE_STATE_CONNECTING;
    _init_step = 0;
    _sequence = 1;

    // Initialise config parameters to similar as the default drone setup
    _control_yaw = 1.75;       // radians/sec
    _control_vz_max = 700.0;   // mm/sec
    _euler_angle_max = 0.1;    // Radians
    _altitude_max = 3000.0;    // mm
    _outdoor = false;

    // Initialise flying parameters to something safe
    _roll = 0.0;
    _pitch = 0.0;
    _gaz = 0.0;
    _yaw = 0.0;

    // Clear the command buffer
    cmd[0] = 0;
}

/////////////////////////////////////////////////////////////////////
void Ardrone::init()
{
    WiFi.init();
}

/////////////////////////////////////////////////////////////////////
void Ardrone::run()
{
    WiFi.run();
}

/////////////////////////////////////////////////////////////////////
// AVR sprintf in Arduino cannot format floats
// This function formats floats to 2 decimal places
// Caution: stores result in static string.
static const char* formatFloat(float f)
{
    static char s[20];
    return dtostrf(f, 1, 2, s);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtPCmd(unsigned int flags, float roll, float pitch, float gaz, float yaw)
{
    sprintf(cmd + strlen(cmd), "AT*PCMD=%ld,%d,%ld,%ld,%ld,%ld\r", _sequence++, flags, 
	    *(unsigned long*)&roll, *(unsigned long*)&pitch, *(unsigned long*)&gaz, *(unsigned long*)&yaw);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtConfigCmd(const char* name, const char* value)
{
    sprintf(cmd + strlen(cmd), "AT*CONFIG=%ld,\"%s\",\"%s\"\r", _sequence++, name, value);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtLedCmd(Ardrone::LEDAnimation animation, float frequency, uint16_t duration)
{
    sprintf(cmd + strlen(cmd), "AT*LED=%ld,%d,%ld,%d\r", _sequence++, animation, *(unsigned long*)&frequency, duration);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtAnimCmd(Ardrone::Animation animation, uint16_t duration)
{
    sprintf(cmd + strlen(cmd), "AT*ANIM=%ld,%d,%d\r", _sequence++, animation, duration);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtFtrimCmd()
{
    sprintf(cmd + strlen(cmd), "AT*FTRIM=%ld\r", _sequence++);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtRefCmd(uint32_t flags)
{
    sprintf(cmd + strlen(cmd), "AT*REF=%ld,%ld\r", _sequence++, flags);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtComWdgCmd()
{
    sprintf(cmd + strlen(cmd), "AT*COMWDG=%ld\r", _sequence++);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtPmodeCmd(uint8_t mode)
{
    sprintf(cmd + strlen(cmd), "AT*PMODE=%ld,%d\r", _sequence++, mode);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::addAtMiscCmd(uint16_t a, uint16_t b, uint16_t c, uint16_t d)
{
    sprintf(cmd + strlen(cmd), "AT*MISC=%ld,%d,%d,%d,%d\r", _sequence++, a, b, c, d);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::sendAtCmd(uint8_t* msg, uint16_t len)
{
    uip_send(msg, len);
}

/////////////////////////////////////////////////////////////////////
void Ardrone::sendAtCmd(char* s)
{
    sendAtCmd((uint8_t*)s, strlen(s));
}

/////////////////////////////////////////////////////////////////////
void Ardrone::sendCurrentAtCmd()
{
    sendAtCmd(cmd);
    cmd[0] = 0;
}

/////////////////////////////////////////////////////////////////////
// Update the drone with the current flying input settings
void Ardrone::update()
{
    char cmd[200];
    uint32_t pcmd_flags = ARDRONE_PCMD_FLAG_PROGRESSIVE;
    uint32_t ref_flags = ARDRONE_REF_FLAG_BASIC;

    if (_flying)
    {
	ref_flags |= ARDRONE_REF_FLAG_START;
	addAtPCmd(pcmd_flags, _roll, _pitch, _gaz, _yaw);
	addAtComWdgCmd();
    }
    addAtRefCmd(ref_flags);
    sendCurrentAtCmd();
}

/////////////////////////////////////////////////////////////////////
// Called when a TCP connection has been made to the ARDRONE, which means we are associated 
// and ARP is done. Can now continute with initialising the drone
void Ardrone::connected()
{
    _state = ARDRONE_STATE_INITIALISING;
    _init_step = 0;
}

/////////////////////////////////////////////////////////////////////
// Initialise the AR.Drone with a sequence of initialisation commands
// This is called after we are pretty sure we are associated and ARP has been done
// The steps are throttled to be sure to allow the drone to get each one
void Ardrone::device_init()
{
    switch (_init_step++)
    {
	case 10:
	    addAtConfigCmd("control:outdoor", _outdoor ? "TRUE" : "FALSE");
	    sendCurrentAtCmd();
	    break;

	case 20:
	    addAtConfigCmd("control:control_yaw", formatFloat(_control_yaw));
	    sendCurrentAtCmd();
	    break;

	case 30:
	    addAtConfigCmd("control:control_vz_max", formatFloat(_control_vz_max));
	    sendCurrentAtCmd();
	    break;

	case 40:
	    addAtConfigCmd("control:euler_angle_max", formatFloat(_euler_angle_max));
	    sendCurrentAtCmd();
	    break;

	case 50:
	    addAtConfigCmd("control:altitude_max", formatFloat(_altitude_max));
	    sendCurrentAtCmd();
	    break;

	case 60:
	    addAtConfigCmd("control_level", "1");
	    sendCurrentAtCmd();
	    break;

	case 70:
	    addAtFtrimCmd();
	    sendCurrentAtCmd();
	    break;

	case 80:
	    addAtLedCmd(ARDRONE_LED_ANIMATION_BLINK_GREEN_RED, 5.0, 5);
	    sendCurrentAtCmd();
	    break;

	case 90:
	    _state = ARDRONE_STATE_INITIALISED;
	    break;
    }
}

/////////////////////////////////////////////////////////////////////
// Periodically attend to the drone
void Ardrone::periodicTask()
{
    unsigned long current_t = millis();
    static unsigned long next_t = 0;
    if (current_t > next_t)
    {
	if (_state == ARDRONE_STATE_INITIALISING)
	    device_init();
	else if (_state == ARDRONE_STATE_INITIALISED)
	    update();
	next_t = current_t + ARDRONE_AT_CMD_INTERVAL;
    }
}

/////////////////////////////////////////////////////////////////////
// SETTERS for initialisation parameters

void Ardrone::setControlYaw(float control_yaw)
{
    _control_yaw = control_yaw;
}

void Ardrone::setControlVZMax(float control_vz_max)
{
    _control_vz_max = control_vz_max;
}

void Ardrone::setEulerAngleMax(float euler_angle_max)
{
    _euler_angle_max = euler_angle_max;
}

void Ardrone::setAlitudeMax(float altitude_max)
{
    _altitude_max = altitude_max;
}

void Ardrone::setOutdoor(float outdoor)
{
    _outdoor = outdoor;
}

/////////////////////////////////////////////////////////////////////
// ACCESSORS for flying parameters

void Ardrone::setFlying(boolean flying)
{
    _flying = flying;
}
boolean Ardrone::flying()
{
    return _flying;
}

void    Ardrone::setRoll(float roll)
{
    _roll = roll;
}
float   Ardrone::roll()
{
    return _roll;
}

void    Ardrone::setPitch(float pitch)
{
    _pitch = pitch;
}
float   Ardrone::pitch()
{
    return _pitch;
}

void    Ardrone::setGaz(float gaz)
{
    _gaz = gaz;
}
float   Ardrone::gaz()
{
    return _gaz;
}

void    Ardrone::setYaw(float yaw)
{
    _yaw = yaw;
}
float   Ardrone::yaw()
{
    return _yaw;
}


